#include "swan.h"

#define SW_TR SWPOINT
#define DL_TYPE DL_Point

///////////////////////////////////////////////////////////////////////
//
DL_TYPE::DL_TYPE()
//
// Purpose: initializes a linked list object.
//
{
    nodeCount = 0;
    nodePtr = NULL;
    headPtr = NULL;
    tailPtr = NULL;
    errMsg[0] = '\0';
}

//////////////////////////////////////////////////////////////////////
//
DL_TYPE& DL_TYPE::copy(DL_TYPE&  dl)
//
// Purpose: copies a generic doubly-linked linked.
//
// Paramaters:
//
//    input: dl - the list to be copied.
//
{
    ListNode *p, *q, *r;

    if (headPtr)
        clear();
    nodeCount = dl.nodeCount;
    nodePtr = NULL;
    headPtr = NULL;
    tailPtr = NULL;
    errMsg[0] = '\0';
    if (!dl.headPtr)
        return *this;
    headPtr = new ListNode;
    if (!headPtr) {
        strcpy(errMsg, ALLOCATE_ERROR);
        return *this;
    }
    headPtr->listNodeData = dl.headPtr->listNodeData;
    headPtr->nextNode = NULL;
    headPtr->prevNode = NULL;
    tailPtr = headPtr;
    p = headPtr;
    r = NULL;
    q = dl.headPtr->nextNode;
    while (q) {
        r = new ListNode;
        if (!r) {
            strcpy(errMsg, ALLOCATE_ERROR);
            return *this;
        }
        r->listNodeData = q->listNodeData;
        r->nextNode = NULL;
        r->prevNode = p;
        p->nextNode = r;
        tailPtr = r;
        p = p->nextNode;
        q = q->nextNode;
    }
    return *this;
}

////////////////////////////////////////////////////////////////////
//
char* DL_TYPE::getErrMsg()
//
// Purpose: returns the contents of the errMsg data member and
// assigns a null string to the errMsg member.
//
{
   return errMsg ;
}

////////////////////////////////////////////////////////////////////
//
bool DL_TYPE::insert(SW_TR& x)
//
// Purpose: performs unordered list insertion.  each new
// element becomes the last on in the list.
//
// Paramaters:
//
//    input: x - the data item inserted in the list.
//
{
    errMsg[0] = '\0';
    if (!headPtr) {
        headPtr = new ListNode;
        if (!headPtr) {
          strcpy(errMsg, ALLOCATE_ERROR);
          return false;
        }
        headPtr->listNodeData = x;
        headPtr->nextNode = NULL;
        headPtr->prevNode = NULL;
        tailPtr = headPtr;
    }
    else {
        thisPtr = new ListNode;
        if (!thisPtr) {
          strcpy(errMsg, ALLOCATE_ERROR);
          return false;
        }
        thisPtr->listNodeData = x;
        thisPtr->prevNode = tailPtr;
        thisPtr->nextNode = NULL;
        tailPtr->nextNode = thisPtr;
        tailPtr = thisPtr;
    }
    nodeCount++;
    return true;
}

////////////////////////////////////////////////////////////////////
//
bool DL_TYPE::insert(SW_TR& x, SW_TR& y)
//
// Purpose: performs unordered list insertion.  
// element y is next to element x in the list.
//
// Paramaters:
//
//    input: x - the data item to specify the insert position.
//    input: y - the data item to be inserted after element x.
//
{
    errMsg[0] = '\0';
    if (!headPtr) {
        strcpy(errMsg, ALLOCATE_ERROR);
        return false;
    }
    else {
        if (search(x, 1)) { // x is found  
	    nodePtr = new ListNode ;
            if (!nodePtr) {
          	strcpy(errMsg, ALLOCATE_ERROR);
          	return false;
            }
            nodePtr->listNodeData = y;
            nodePtr->prevNode = thisPtr;
            nodePtr->nextNode = thisPtr->nextNode ;
	    if (thisPtr->nextNode != NULL) { // succesor exist
		thisPtr->nextNode->prevNode = nodePtr ;
		thisPtr->nextNode = nodePtr ;
	    } 
	    else { // no successor
		thisPtr->nextNode = nodePtr ;	
                tailPtr = nodePtr;
	    }
	}
	else 
	   return false ; 
    }
    nodeCount++;
    return true;
}
///////////////////////////////////////////////////////////////////
//
bool DL_TYPE::search(SW_TR& x, unsigned occur)
//
// Purpose: searches for x in the unordered list starting at the
// head of the list.  The pointer to the current node is returned
// in the argument list.  The complete datum of the matching node
// is accessed by parameter x (this is relevant only for structure
// data).
//
// Paramaters:
//
//    input: x - the sought data element.
//           occur - the sought occurrence of a search key.
//
//    output: thisPtr - the pointer to the sought list element.
//
{

    occur = (occur > 0) ? occur : 1;
    thisPtr = headPtr;

    while (thisPtr && occur > 0) {
        if (x == thisPtr->listNodeData)
            occur--;
        if (occur > 0)
            thisPtr = thisPtr->nextNode;
    }

    if (occur == 0) // copy data to pointer x
        x = thisPtr->listNodeData;

    return (occur == 0) ? true : false;
}

//////////////////////////////////////////////////////////////////////
//
bool DL_TYPE::remove(SW_TR& x, unsigned occur)
//
// Purpose: attempts to delete an item from an ordered list.
// The boolean function result confirms the deletion process.
//
// Paramaters:
//
//    input: x - the data element to be deleted from the list.
//           occur - the sought occurrence of a search key.
//
{
    ListNode *p;
    bool match;

    occur = (occur > 0) ? occur : 1;
    // search for item to delete
    match = search(x, occur);

    if (match) {
        if (nodeCount == 1) {
            delete headPtr;
            headPtr = NULL;
            tailPtr = NULL;
        }
        else if (!thisPtr->nextNode) {
            // delete the last list member
            p = thisPtr->prevNode;
            p->nextNode = NULL;
            delete tailPtr;
            // link tailPtr pointer with new last element
            tailPtr = p;
        }
        else if (!thisPtr->prevNode) {
            // delete the list headPtr
            p = thisPtr->nextNode;
            p->prevNode = NULL;
            delete headPtr;
            headPtr = p; // link headPtr with new first element
        }
        else {
            thisPtr->prevNode->nextNode = thisPtr->nextNode;
            thisPtr->nextNode->prevNode = thisPtr->prevNode;
            delete thisPtr;
        }
        nodeCount--;
    }
    return match; // return function value
}

///////////////////////////////////////////////////////////////////////
//
bool DL_TYPE::getFirstNode(SW_TR& x)
//
// Purpose: visits the first node of the generic linked list.  The
// function returns true if the list is not empty.  Otherwise it
//  yields false.
//
// Paramaters:
//
//    output: x - the data obtained from the first list node.
//
{
    if (headPtr) {
        x = headPtr->listNodeData;
        nodePtr = headPtr;
        return true;
    }
    else
        return false;
}

/////////////////////////////////////////////////////////////////////
//
bool DL_TYPE::getLastNode(SW_TR& x)
//
// Purpose: visits the tail of the linked list.  The function yields
// true if the list is not empty.  Otherwise, the function returns
// false.
//
// Paramaters:
//
//    output: x - the obtained from the last list node.
//
{
    if (tailPtr) {
        x = tailPtr->listNodeData;
        nodePtr = tailPtr;
        return true;
    }
    else
        return false;
}

/////////////////////////////////////////////////////////////////////
//
bool DL_TYPE::getNextNode(SW_TR& x)
//
// Purpose: visits the next node in the generic list.  The function
// returns false if an attempt is made to read beyond the tail of
// the list.  Otherwise, the function yields true.
//
// Paramaters:
//
//    output: x - the data obtained from the next list node.
//
{
    if (nodePtr && nodePtr->nextNode) {
        nodePtr = nodePtr->nextNode;
        x = nodePtr->listNodeData;
        return true;
    }
    else
        return false;
}

/////////////////////////////////////////////////////////////////////
//
bool DL_TYPE::getNextNode(SW_TR x, SW_TR& y)
//
// Purpose: visits the next node of x in the generic list.  The function
// returns false if an attempt is made to read beyond the tail of
// the list.  Otherwise, the function yields true.
//
// Paramaters:
//
//    input: x - the node to be checked
//    output: y - the next node of x 
//	      
{
    if (search(x, 1)) { // x exist
	if (thisPtr->nextNode != NULL) {
	    y = thisPtr->nextNode->listNodeData ;
	    nodePtr = thisPtr ;
	    return true ;
	}
    }
    return false ;
}
	
/////////////////////////////////////////////////////////////////////
//
bool DL_TYPE::getPrevNode(SW_TR& x)
//
// Purpose: visits the previous node in the generic list.  The
// function returns false if an attempt is made to read beyond
// the tail of the list.  Otherwise, yields true.
//
// Paramaters:
//
//    output: x - the data obtained from the previous list node.
//
{
    if (nodePtr && nodePtr->prevNode) {
        nodePtr = nodePtr->prevNode;
        x = nodePtr->listNodeData;
        return true;
    }
    else
        return false;
}

/////////////////////////////////////////////////////////////////////
//
bool DL_TYPE::getPrevNode(SW_TR x, SW_TR& y)
//
// Purpose: visits the previous node in the generic list.  The
// function returns false if an attempt is made to read beyond
// the tail of the list.  Otherwise, yields true.
//
// Paramaters:
//
//    output: x - the data obtained from the previous list node.
//
{
    if (search(x, 1)) { // x exist
	if (thisPtr->prevNode != NULL) {
	    y = thisPtr->prevNode->listNodeData ;
	    nodePtr = thisPtr ;
	    return true ;
	}
    }
    return false ;
}
	
//////////////////////////////////////////////////////////////////////
//
bool DL_TYPE::getData(SW_TR& x, 
                               unsigned index, 
                               bool updateNodePtr)
//
// Purpose: obtains the data at node number index.  The function
// returns true if index is less than the nodeCount member.
//
// Parameters:
//
//    input: index - the index of the sought node
//           updateNodePtr - flag to update nodePtr member.
//
//    output: x - the data obtained from the targeted list node.
//
{
  ListNode *p;

  if (index >= nodeCount)
    return false;

  p = headPtr;
  // search for the node number index
  while (index > 0) {
    p = p->nextNode;
    index--;
  }
  x = p->listNodeData;
  nodePtr = (updateNodePtr) ? p : nodePtr;
  return true;
}

/////////////////////////////////////////////////////////////////////
//
void DL_TYPE::clear()
//
// Purpose: clears the linked list.
//
{
    ListNode* p;

    while (headPtr) {
        p = headPtr;
        headPtr = headPtr->nextNode;
        delete p;
        nodeCount--;
    }
    tailPtr = NULL;
    nodePtr = NULL;
}


#undef SW_TR
#undef DL_TYPE
