/************************************************************
 * edge.c --- member functions of class BasicEdge	    *
 * Jun Yang, November 17, 1994				    *
 ************************************************************/

/*--------------- Header Files ---------------*/

#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
extern "C" {
#include "geosim.h"
#include "gspriv.h"
}
#include "swan.h"

#define DELTAX	3
#define DELTAY  3
#define DEFARROWLENCOSA	9.66
#define DEFARROWLENSINA 2.58

/*--------------- Member Functions of BasicEdge ---------------*/

// constructors and destructor

BasicEdge::BasicEdge(INNODEID src_node, INNODEID dest_node, LABEL name)
{
	s_node = src_node ;
	e_node = dest_node ;
	if (name != NULL) {
            label = new char[strlen(name)+1] ;
            strcpy(label, name) ;
	}
	else
	    label = NULL ;
	rpath = new CURVE ;
	ppath = new CURVE ;
	wpath = new CURVE ;
	active = FALSE ;
	info = new EDGEINFO ;
}

BasicEdge::BasicEdge()
{
	label = NULL ;
	rpath = new CURVE ;
	ppath = new CURVE ;
	wpath = new CURVE ;
	active = FALSE ;
	info = new EDGEINFO ;
}

BasicEdge::~BasicEdge(void)
{
	delete info ;
	delete wpath ;
	delete ppath ;
	delete rpath ;

	info = NULL ;
	wpath = NULL ;
	ppath = NULL ;
	rpath = NULL ;
}

// member access functions 

bool BasicEdge::setLayoutComp(LAYOUTCOMPID layout_id)
{
	lc_id = layout_id ;
	return TRUE ;
}

LAYOUTCOMPID BasicEdge::getLayoutComp(void)
{
	return lc_id ;
}

bool BasicEdge::setAttr(EDGEINFO* attrs)
{
	*info = *attrs ;
	return TRUE ;
}

BasicNode* BasicEdge::getSNode(void)
{
	return s_node ;
}

BasicNode* BasicEdge::getENode(void)
{
	return e_node ;
}

LABEL BasicEdge::getLabel(void)
{
	return label ;
}

bool BasicEdge::setLabel(LABEL new_label)
{
        if (new_label != NULL) {
            if (label != NULL)
                delete[] label ;
            label = new char[strlen(new_label)+1] ;
            strcpy(label, new_label) ;
        }
        else {
            if (label != NULL)
                delete[] label ;
            label = NULL ;
        }
        return TRUE ;
}

EDGEINFO* BasicEdge::getAttr(void)
{
	return info ;
}

bool BasicEdge::setRPath(CURVE* r_path)
{
	SWPOINT	p1 ;
	bool	ok ;

	(*rpath).clear() ;
	ok = (*r_path).getFirstNode(p1) ;
	while (ok) {
	    (*rpath).insert(p1) ;
	    ok = (*r_path).getNextNode(p1) ;
	}
	return TRUE ;
}

CURVE* BasicEdge::getRPath(void)
{
	return rpath ;
}

bool BasicEdge::setPPath(CURVE* p_path)
{
	SWPOINT	p1 ;
	bool	ok ;

	(*ppath).clear() ;
	ok = (*p_path).getFirstNode(p1) ;
	while (ok) {
	    (*ppath).insert(p1) ;
	    ok = (*p_path).getNextNode(p1) ;
	}
	return TRUE ;
}

CURVE* BasicEdge::getPPath(void)
{
	return ppath ;
}

bool BasicEdge::setWPath(CURVE* w_path)
{
	SWPOINT	p1 ;
	bool	ok ;

	(*wpath).clear() ;
	ok = (*w_path).getFirstNode(p1) ;
	while (ok) {
	    (*wpath).insert(p1) ;
	    ok = (*w_path).getNextNode(p1) ;
	}
	return TRUE ;
}

CURVE* BasicEdge::getWPath(void)
{
	return wpath ;
}

bool BasicEdge::setColor(COLORTYPE color) 
{
	info->color = color ;
	return TRUE ;
}

COLORTYPE BasicEdge::getColor(void) 
{
	return info->color ;
}

bool BasicEdge::setLineTH(int line_TH) 
{
	info->lineTH = line_TH ;
	return TRUE ;
}

int BasicEdge::getLineTH(void) 
{
	return info->lineTH ;
}

bool BasicEdge::setLength(int e_length) 
{
	info->length = e_length ;
	return TRUE ;
}

int BasicEdge::getLength(void) 
{
	return info->length ;
}

bool BasicEdge::setArrow(ARROWTYPE e_arrow) 
{
	info->arrow_type = e_arrow ;
	return TRUE ;
}

void BasicEdge::enActive(void)
{
	active = TRUE ;
}

void BasicEdge::disActive(void)
{
	active = FALSE ;
}

FLAG BasicEdge::isDisplayed()
{
	return active ;
}

bool BasicEdge::calWPath(void)
{
	LOCATION	pos1 ;
	bool		ok ;

	(*wpath).clear() ;
	ok = (*ppath).getFirstNode(pos1) ;
	while (ok) {
	    pos1.x = (int)(pos1.x * zoom_factor) - sc_x_origin ;
	    pos1.y = (int)(pos1.y * zoom_factor) - sc_y_origin ;
	    (*wpath).insert(pos1) ;
	    ok = (*ppath).getNextNode(pos1) ; 
	}
	return TRUE ;
}

bool BasicEdge::calLabelWPos(void)
{
	int		len ;

	if (label != NULL)
	    len = CURFONTWIDTH * strlen(label) ;
	switch (info->label_orient) {
	case HORIZONTAL:
	    info->label_wpos.x = (int)((info->label_ppos.x+len/2)
			         * zoom_factor) - sc_x_origin - len/2 ;
	    info->label_wpos.y = (int)(info->label_ppos.y * zoom_factor) -
				sc_y_origin ;
	    break ;

	case VERTICAL:
	case VICEDIAGONAL:
	    info->label_wpos.x = (int)((info->label_ppos.x+len)
			         * zoom_factor) - sc_x_origin - len ;
	    info->label_wpos.y = (int)(info->label_ppos.y * zoom_factor) -
				sc_y_origin ;
	    break ;

	case DIAGONAL:
	    info->label_wpos.x = (int)((info->label_ppos.x+len)
			         * zoom_factor) - sc_x_origin - len ;
	    info->label_wpos.y = (int)((info->label_ppos.y-CURFONTHEIGHT)
				 * zoom_factor) - sc_y_origin 
				 + CURFONTHEIGHT ;
	    break ;
	}
	return TRUE ;
}
	
bool BasicEdge::_intersections(LOCATION s_pos, NODETYPE s_type,
		BOXSIZE sp, LOCATION e_pos, NODETYPE e_type,
		BOXSIZE ep, LOCATION& i_pos1, LOCATION& i_pos2)
{
	LOCATION c1_pos, c2_pos ;

	switch (s_type) {
	case BOX:
	    c1_pos.x = s_pos.x + sp.w / 2 ; 
	    c1_pos.y = s_pos.y + sp.h / 2 ;
	    break ;
	case CIRCLE:
	    c1_pos.x = s_pos.x + sp.w / 2 ; 
	    c1_pos.y = s_pos.y + sp.h / 2 ;
	    break ;
	default:
	    return FALSE ;
	}

	switch (e_type) {
	case BOX:
	    c2_pos.x = e_pos.x + ep.w / 2 ; 
	    c2_pos.y = e_pos.y + ep.h / 2 ;
	    break ;
	case CIRCLE:
	    c2_pos.x = e_pos.x + ep.w / 2 ; 
	    c2_pos.y = e_pos.y + ep.h / 2 ;
	    break ;
	default:
	    return FALSE ;
	}

	
	// calculate cross point near s node
	switch (s_type) {
	case BOX:
	    line_box_crosspoint(c2_pos, c1_pos, sp.w, sp.h, i_pos1) ;
	    break ;
	case CIRCLE:
	    line_ellipse_crosspoint(c2_pos, s_pos, sp.w, sp.h, i_pos1) ;
	    break ;
	}

	// calculate cross point near e node
	switch (e_type) {
	case BOX:
	    line_box_crosspoint(c1_pos, c2_pos, ep.w, ep.h, i_pos2) ;
	    break ;
	case CIRCLE:
	    line_ellipse_crosspoint(c1_pos, e_pos, ep.w, ep.h, i_pos2) ;
	    break ;
	}
	return TRUE ;
}

bool BasicEdge::getLast2Nodes(LOCATION& pos1, LOCATION& pos2)
{
	LOCATION	temp_pos1, temp_pos2, temp_pos3 ;
	bool		ok ;

	if ((*wpath).getNodeCount() < 2)
	    return FALSE ;
	(*wpath).getFirstNode(temp_pos1) ;
	(*wpath).getNextNode(temp_pos2) ;
	ok = (*wpath).getNextNode(temp_pos3) ;
	while (ok) {
	    temp_pos1 = temp_pos2 ;
	    temp_pos2 = temp_pos3 ;
	    ok = (*wpath).getNextNode(temp_pos3) ;
	}
	pos1 = temp_pos1 ;
	pos2 = temp_pos2 ;
	return TRUE ;
}
	     
bool BasicEdge::getLast2PPath(LOCATION& pos1, LOCATION& pos2)
{
	LOCATION	temp_pos1, temp_pos2, temp_pos3 ;
	bool		ok ;

	if ((*ppath).getNodeCount() < 2)
	    return FALSE ;
	(*ppath).getFirstNode(temp_pos1) ;
	(*ppath).getNextNode(temp_pos2) ;
	ok = (*ppath).getNextNode(temp_pos3) ;
	while (ok) {
	    temp_pos1 = temp_pos2 ;
	    temp_pos2 = temp_pos3 ;
	    ok = (*ppath).getNextNode(temp_pos3) ;
	}
	pos1 = temp_pos1 ;
	pos2 = temp_pos2 ;
	return TRUE ;
}

bool BasicEdge::getFirst2PPath(LOCATION& pos1, LOCATION& pos2)
{
	if ((*ppath).getNodeCount() < 2)
	    return FALSE ;
	(*ppath).getFirstNode(pos1) ;
	(*ppath).getNextNode(pos2) ;
	return TRUE ;
}

// pick an edge 

bool BasicEdge::nearThisEdge(int x, int y)
{
	SWPOINT	p1, p2 ;
	int	min_x, min_y, max_x, max_y ;
	double	a2, b2, c2 ;
	bool	ok ;
	double  delta ;

	if (!isDisplayed())
	    return FALSE ;
	
	ok = (*wpath).getFirstNode(p1) ;
	if (ok) {
	    ok = (*wpath).getNextNode(p2) ;
	    while (ok) {
		if (p1.x <= p2.x) {
		    min_x = p1.x - DELTAX ;
		    max_x = p2.x + DELTAX ;
		}
		else {
		    min_x = p2.x - DELTAX ;
		    max_x = p1.x + DELTAX ;
		}
		if (p1.y <= p2.y) {
		    min_y = p1.y - DELTAY ;
		    max_y = p2.y + DELTAY ;
		}
		else {
		    min_y = p2.y - DELTAY ;
		    max_y = p1.y + DELTAY ;
		}
		if ((x >= min_x) && (x <= max_x) && 
		    (y >= min_y) && (y <= max_y)) {	// within the area
		    a2 = (double)((long)(x - p1.x)*(long)(x - p1.x) + 
			 (long)(y - p1.y)*(long)(y - p1.y)) ;
		    b2 = (double)((long)(x - p2.x)*(long)(x - p2.x) + 
			 (long)(y - p2.y)*(long)(y - p2.y)) ;
		    c2 = (double)((long)(p1.x - p2.x)*(long)(p1.x - p2.x) + 
			 (long)(p1.y - p2.y)*(long)(p1.y - p2.y)) ;

		    delta = fabs ((double)(a2 + c2 - b2) /
                                (2.0 * sqrt(a2 * c2)) - 1.0) ;
		    if (delta < 0.005) {
			_sw_pick_edge = this ;
        		_sw_pick_graph = (*(*_sw_pick_edge).getLayoutComp()).getGID() ;
			return TRUE ;
		    }
		}
		p1 = p2 ;
		ok = (*wpath).getNextNode(p2) ;
	    }
	}
	return FALSE ;
}

// Input/Output methods 
//	. save
// 	. load	

bool BasicEdge::save(ofstream& edge_out)
{
	LOCATION	pos ;
	bool		ok ;

	if (edge_out == NULL)
	    return FALSE ;

	edge_out << '(' << (*lc_id).getEID() << ' ' << (int)(*s_node).getID()
		 << ' ' << (int)(*e_node).getID() << ' ' 
	         << '\"' ;
	if (label != NULL)
	    edge_out << label ;
	edge_out << '\"' << ' '
		 << info->lineTH << ' ' << info->color << ' '
		 << info->label_ppos.x << ' ' << info->label_ppos.y 
		 << ' ' << (*ppath).getNodeCount() << ')' << '\n' ;
	
	edge_out << '[' << ' ' ;
	ok = (*ppath).getFirstNode(pos) ;
	while (ok) {
	    edge_out << pos.x << ' ' << pos.y << ' ' ;
	    ok = (*ppath).getNextNode(pos) ;
	}
	edge_out << ']' << '\n' ;
	return TRUE ;
}

bool BasicEdge::load(ifstream& edge_in, BasicGraph* parent)
{
	char 		c, *p ;
	BasicNode 	*node1, *node2 ;
	NODEID		node1_id, node2_id ;
	int		temp_node1, temp_node2 ;
	LC_ID		temp_lceid ;
	int		i, num_points, temp_len ;
	BasicLayoutComp *lc1, *lc2 ;
	LOCATION	temp_pos ;
	char		temp_str[80] ;

	if (edge_in == NULL)
	    return FALSE ;
	edge_in >> c ;
	if (c != '(')
	    return FALSE ;
	edge_in >> temp_lceid >> temp_node1 >> temp_node2 >> c ;
	node1_id = (NODEID)temp_node1 ;
	node2_id = (NODEID)temp_node2 ;

	if (c != '\"')
	    return FALSE ;
	p = temp_str ;
	edge_in >> c ;
	while (c != '\"') {
	    *p++ = c ;
	    edge_in >> c ;
	}
	*p = '\0' ;
	temp_len = strlen(temp_str) ;
	if (label == NULL) 
	    label = new char[temp_len+1] ;
	else if (strlen(label) < temp_len) {
	    delete[] label ;
	    label = new char[temp_len+1] ;
	}
	strcpy(label, temp_str) ;

	edge_in >> info->lineTH >> info->color >> info->label_ppos.x 
	        >> info->label_ppos.y >> num_points >> c ;
	if (c != ')')
	    return FALSE ;

	if ((s_node = (*parent).searchNode(node1_id)) == NULL)
	    return FALSE ;

	if ((e_node = (*parent).searchNode(node2_id)) == NULL)
	    return FALSE ;
	
	lc1 = (*s_node).getLayoutComp() ;
	lc2 = (*e_node).getLayoutComp() ;
	if (lc1 == lc2)
	    lc_id = lc1 ;
	else 
	    lc_id = (*parent).findCCA(lc1, lc2) ;

	(*ppath).clear() ;
	edge_in >> c ;
	if (c != '[')
	    return FALSE ;
	for (i = 0 ; i < num_points ; i++) {
	    edge_in >> temp_pos.x >> temp_pos.y ;
	    (*ppath).insert(temp_pos) ;
	}
	edge_in >> c ;
	if (c != ']') 
	    return FALSE ;

	return TRUE ;
}

// display an edge 

bool BasicEdge::displayEdge()
{
	active = TRUE ;
	calWPath() ;
	calLabelWPos() ;
	swg_setcurrentcolor(info->color) ;
	swg_setcurrentsize(info->lineTH) ;
        swg_drawcurve(wpath) ;
	displayArrow() ;
	if (((*(*lc_id).getGID()).getGraphInfo())->fShowLabel) 
	    displayLabel() ;
	return TRUE ;
}

bool BasicEdge::erase()
{
	swg_clearpath(wpath) ;
	return TRUE ;
}

bool BasicEdge::dump()
{
	if (!isDisplayed()) 
	    return FALSE ;

	swg_setcurrentcolor(info->color) ;
	swg_setcurrentsize(info->lineTH) ;
        swg_drawcurve(wpath) ;
	displayArrow() ;
	if (((*(*lc_id).getGID()).getGraphInfo())->fShowLabel) 
	    displayLabel() ;
	return TRUE ;
}

bool BasicEdge::displayArrow(void)
{
	LOCATION	pos1, pos2, pos3 ;
	double		angleB, cosB, sinB ;

	switch (info->arrow_type) {
	case NOARROW:
	    break ;

	case ENDARROW:
	    if (getLast2Nodes(pos1, pos2)) {
		if (pos1.x == pos2.x) { // vertical line
		    pos3.x = pos1.x - 2 ;
		    if (pos1.y > pos2.y) {
			pos3.y = pos2.y + 4 ;
			swg_drawline(pos3, pos2) ;
			pos3.x = pos1.x + 2 ;
			swg_drawline(pos3, pos2) ;
		    }
		    else {
			pos3.y = pos2.y - 4 ;
			swg_drawline(pos3, pos2) ;
			pos3.x = pos1.x + 2 ;
			swg_drawline(pos3, pos2) ;
		    }
		}
		else if (pos1.y == pos2.y) { // horizontal line
		    pos3.y = pos1.y - 2 ;
                    if (pos1.x > pos2.x) {
                        pos3.x = pos2.x + 4 ;
                        swg_drawline(pos3, pos2) ;
                        pos3.y = pos1.y + 2 ;
                        swg_drawline(pos3, pos2) ;
                    }
                    else {
                        pos3.x = pos2.x - 4 ;
                        swg_drawline(pos3, pos2) ;
                        pos3.y = pos1.y + 2 ;
                        swg_drawline(pos3, pos2) ;
                    }
                }
		else { // slant line
		    angleB = atan((double)(pos1.y-pos2.y)/
				  (double)(pos2.x-pos1.x)) ;
		    if (pos2.x > pos1.x) 
			angleB += M_PI ;
		    cosB = cos(angleB) ;
		    sinB = sin(angleB) ;
		    pos3.x = (int) (DEFARROWLENCOSA * cosB - 
			   DEFARROWLENSINA * sinB) + pos2.x ;
		    pos3.y = - (int) (DEFARROWLENCOSA * sinB +
			   DEFARROWLENSINA * cosB) + pos2.y ;
		    swg_drawline(pos3, pos2) ;
		    pos3.x = (int) (DEFARROWLENCOSA * cosB + 
			   DEFARROWLENSINA * sinB) + pos2.x ;
		    pos3.y = - (int) (DEFARROWLENCOSA * sinB -
			   DEFARROWLENSINA * cosB) + pos2.y ;
		    swg_drawline(pos3, pos2) ;
		} // end of slant line
	    }
	    break ;

	case STARTARROW:
	case BOTHARROW:
	    break ;
	}
	return TRUE ;
}
	
void BasicEdge::displayLabel(void)
{
	LOCATION	pos1, pos2 ;
	int		label_len ;

	if (zoom_factor <= ZOOMOUTSTR) 
	    return ;
	if (label == NULL) 
	    return ;

	GSsetfont(SMALL) ;
	switch (lc_id->type) {
	case LISTACROSS:
	    label_len = strlen(label) * CURFONTWIDTH ;
	    if ((*wpath).getNodeCount() == 2) {
		(*wpath).getFirstNode(pos1) ;
		(*wpath).getNextNode(pos2) ;
		if (abs(pos2.x - pos1.x) >= label_len) 
		    swg_drawstring((pos1.x+pos2.x-label_len)/2,
				 pos1.y, label) ;		
	    } 
	    break ;

	case LISTDOWN:
	    label_len = strlen(label) * CURFONTWIDTH ;
	    if ((*wpath).getNodeCount() == 2) {
		if (s_node->wsize.w/2 >= label_len) {
	            (*wpath).getFirstNode(pos1) ;
                    (*wpath).getNextNode(pos2) ;
		    swg_drawstring(pos1.x - label_len, 
				 (pos1.y + pos2.y + CURFONTHEIGHT)/2,
				 label) ;
	        }
	    }
	    break ;

	case TREE:
            swg_setcurrentcolor(BLACK) ;
	    (*wpath).getLastNode(pos1) ;
	    label_len = strlen(label) * CURFONTWIDTH ;
            swg_drawstring(pos1.x - label_len - 2, 
			   pos1.y - (Y_SEPARATION - CURFONTHEIGHT)/2, 
			   label) ;
	    break ;

	case MANUAL:
	case CIRCLENET:
	case KKNET:
	case HIERACHY:
            swg_setcurrentcolor(BLACK) ;
	    swg_drawstring(info->label_wpos.x, info->label_wpos.y,
			   label) ;
	    break ;

	case ARRAYDOWN:
	case ARRAYACROSS:
	default:
	    break ;
	}
}

bool BasicEdge::pan(int xoffset, int yoffset)
{
	SWPOINT		pos1 ;
	bool		ok ;
	CURVE*		temp_curve = new CURVE ;

	ok = (*wpath).getFirstNode(pos1) ;
	while (ok) {
	     pos1.x += xoffset ;
	     pos1.y += yoffset ;
	     (*temp_curve).insert(pos1) ;
	     ok = (*wpath).getNextNode(pos1) ;
	}
	delete wpath ;
	wpath = temp_curve ;
	info->label_wpos.x += xoffset ;
	info->label_wpos.y += yoffset ;
	return TRUE ;
}

bool BasicEdge::zoom(void)
{
	SWPOINT		pos1 ;
	bool		ok ;
	CURVE*		temp_curve = new CURVE ;

	ok = (*ppath).getFirstNode(pos1) ;
	while (ok) {
	     pos1.x = (int)(pos1.x 
			    * zoom_factor - sc_x_origin) ;
	     pos1.y = (int)(pos1.y
			    * zoom_factor - sc_y_origin) ;
	     (*temp_curve).insert(pos1) ;
	     ok = (*ppath).getNextNode(pos1) ;
	}
	delete wpath ;
	wpath = temp_curve ;
	calLabelWPos() ;
	return TRUE ;
}

/************************************************************
 * End of edge.c 					    *
 ************************************************************/
