/************************************************************
 * geotools.c -- tools for geometric computation in Swan    *
 * Jun Yang, November 22, 1994				    *
 ************************************************************/

/*--------------- Header Files ---------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "swan.h"

/*--------------- Tools ---------------*/

bool line_box_crosspoint(LOCATION ppos, LOCATION cpos,
			    int width, int height, LOCATION& ip_pos)
{
	long	half_w = width / 2 ;
	long	half_h = height / 2 ;
	long	delta_x = ppos.x - cpos.x ;
	long	delta_y = ppos.y - cpos.y ;

	if (delta_x > half_w) { // right side 
	    if (delta_y < -half_h) { // upper right corner
	      if (labs(delta_y) >= labs(delta_x)*half_h/half_w) {
		ip_pos.y = cpos.y - (int)half_h ;
		ip_pos.x = cpos.x - (int) (delta_x * half_h / delta_y) ;
	      }
	      else {
		ip_pos.x = cpos.x + (int)half_w ;
		ip_pos.y = cpos.y + (int) (delta_y * half_w / delta_x) ;
	      }
	    }
	    else if (delta_y > half_h) { // down right corner
	      if (labs(delta_y) >= labs(delta_x)*half_h/half_w) {
		ip_pos.y = cpos.y + (int)half_h ; 
		ip_pos.x = cpos.x + (int) (delta_x * half_h / delta_y) ;
	      }
	      else {
		ip_pos.x = cpos.x + (int)half_w ; 
		ip_pos.y = cpos.y + (int) (delta_y * half_w / delta_x) ;
	      }
	    }
	    else { // right central part
		ip_pos.x = cpos.x + (int)half_w ;
		ip_pos.y = cpos.y + (int)(delta_y*half_w/delta_x) ;  
	    }
	}
	else if (delta_x < -half_w) { // left side
	    if (delta_y < -half_h) { // upper left corner
	      if (labs(delta_y) >= labs(delta_x)*half_h/half_w) {
		ip_pos.y = cpos.y - (int)half_h ;
		ip_pos.x = cpos.x - (int) (delta_x * half_h / delta_y) ;
	      }
	      else {
		ip_pos.x = cpos.x - (int)half_w ;
		ip_pos.y = cpos.y - (int) (delta_y * half_w / delta_x) ;
	      }
	    }
	    else if (delta_y > half_h) { // down left corner
	      if (labs(delta_y) >= labs(delta_x)*half_h/half_w) {
		ip_pos.y = cpos.y + (int)half_h ; 
		ip_pos.x = cpos.x + (int) (delta_x * half_h / delta_y) ;
	      }
	      else {
		ip_pos.x = cpos.x - (int)half_w ; 
		ip_pos.y = cpos.y - (int) (delta_y * half_w / delta_x) ;
	      }
	    }
	    else { // left central part
		ip_pos.x = cpos.x - (int)half_w ;
		ip_pos.y = cpos.y + (int)(delta_y*half_w/delta_x) ;
	    }
	}
	else { // central part 
	    if (delta_y < -half_h) { // upper central part
		ip_pos.y = cpos.y - (int)half_h ;
		ip_pos.x = cpos.x - (int) (delta_x * half_h / delta_y) ;
	    }
	    else if (delta_y > half_h) { // down central part
		ip_pos.y = cpos.y + (int)half_h ; 
		ip_pos.x = cpos.x + (int) (delta_x * half_h / delta_y) ;
	    }
	    else  // point is within the box
		return FALSE ;
	}
	return TRUE ;
} /* end of line_box_intersections */

bool line_ellipse_crosspoint(LOCATION ppos, LOCATION lpos,
				int lw, int lh, LOCATION& ip_pos)
{
        long    offset_x, offset_y ;
        double  k1, alpha, actual_x, actual_y ;
        int     delta_x, delta_y ;
        int     d, incrH, incrD, quadrant ;
	long	sw, sh, sw2, sh2, sw2h2 ;
	LOCATION cpos ;

	sw = (long)lw / 2 ;
	sh = (long)lh / 2 ;
	cpos.x = lpos.x + (int)sw ;
	cpos.y = lpos.y + (int)sh ;

        if (ppos.x == cpos.x) { // vertical line
            ip_pos.x = ppos.x ;
            if (ppos.y < cpos.y - (int)sh) 
                ip_pos.y = cpos.y - (int)sh ;
            else if (ppos.y > cpos.y + (int)sh)
                ip_pos.y = cpos.y + (int)sh ;
	    else 
		return FALSE ;
	    return TRUE ;
        }

        if (ppos.y == cpos.y) { // horizontal line
            ip_pos.y = ppos.y ;
            if (ppos.x < cpos.x - (int)sw) 
                ip_pos.x = cpos.x - (int)sw ;
            else if (ppos.x > cpos.x + (int)sw)
                ip_pos.x = cpos.x + (int)sw ;
	    else 
		return FALSE ;
	    return TRUE ;
        }

	sw2 = sw * sw ;
	sh2 = sh * sh ;
	sw2h2 = sw2 * sh2 ;

	// general situations
	if (ppos.x < cpos.x) {
	    if (ppos.y < cpos.y) {
		quadrant = 2 ;
		delta_x = cpos.x - ppos.x ;
		delta_y = cpos.y - ppos.y ;
	    }
	    else {
		quadrant = 3 ;
		delta_x = cpos.x - ppos.x ;
		delta_y = ppos.y - cpos.y ;
	    }
	}
	else {
	    if (ppos.y < cpos.y) {
		quadrant = 1 ;
		delta_x = ppos.x - cpos.x ;
		delta_y = cpos.y - ppos.y ;
	    }
	    else {
		quadrant = 4 ;
		delta_x = ppos.x - cpos.x ;
		delta_y = ppos.y - cpos.y ;
	    }
	}

	alpha = (double)delta_y/(double)delta_x ;
	k1 = (double)sw * alpha / (double)sh ;
	actual_x = (double)sw / sqrt(1.0 + k1 * k1) ;
	actual_y = actual_x * alpha ;
	offset_x = (long)actual_x ;
	offset_y = (long)actual_y ;
	if ((double)delta_x * (double)delta_x + 
	    (double)delta_y * (double)delta_y <
	    actual_x * actual_x + actual_y * actual_y)
	    return FALSE ;

	if (sh2*(double)offset_x*(double)offset_x +
	    sw2*(double)offset_y*(double)offset_y < sw2h2) {
	    if (delta_x >= delta_y) { // E & NE	
	        d = 2 * delta_y - delta_x ;
		incrH = 2 * delta_y ;
		incrD = 2 * (delta_y - delta_x) ;
		while (TRUE) {
		    if (d <= 0) { // choose E
		        d += incrH ;
		        offset_x += 1 ;
		    }
		    else { // choose NE
		        d += incrD ;
		        offset_x += 1 ;
		        offset_y += 1 ;
		    }
		    if (sh2*(double)offset_x*(double)offset_x +
		        sw2*(double)offset_y*(double)offset_y >= sw2h2) 
		        break ;
		}
	    }
	    else { // N & NE
	        d = delta_y - 2 * delta_x ;
		incrH = 2 * (-delta_x) ;
		incrD = 2 * (delta_y - delta_x) ;
		while (TRUE) {
		    if (d >= 0) { // choose N
			d += incrH ;
			offset_y += 1 ;
		    }
		    else { // choose NE
			d += incrD ;
			offset_x += 1 ;
			offset_y += 1 ;
		    }
		    if (sh2*(double)offset_x*(double)offset_x +
		        sw2*(double)offset_y*(double)offset_y >= sw2h2) 
		        break ;
		}
	    } // end of justification 
	}

	switch (quadrant) {
	case 1:
	    ip_pos.x = cpos.x + offset_x ;
	    ip_pos.y = cpos.y - offset_y ;
	    break ;
	case 2:
	    ip_pos.x = cpos.x - offset_x ;
	    ip_pos.y = cpos.y - offset_y ;
	    break ;
	case 3:
	    ip_pos.x = cpos.x - offset_x ;
	    ip_pos.y = cpos.y + offset_y ;
	    break ;
	case 4:
	    ip_pos.x = cpos.x + offset_x ;
	    ip_pos.y = cpos.y + offset_y ;
	    break ;
	default:
	    return FALSE ;
	}
	return TRUE ;
} /* end of line_ellipse_crosspoint */

/************************************************************
 * End of geotools.c                                        *
 ************************************************************/
